

python -m aquacrop.solution.aeration_stress
python -m aquacrop.solution.water_stress
python -m aquacrop.solution.evap_layer_water_content
python -m aquacrop.solution.root_zone_water
python -m aquacrop.solution.cc_development
python -m aquacrop.solution.update_CCx_CDC
python -m aquacrop.solution.cc_required_time
python -m aquacrop.solution.temperature_stress
python -m aquacrop.solution.HIadj_pre_anthesis                 
python -m aquacrop.solution.HIadj_post_anthesis
python -m aquacrop.solution.HIadj_pollination
python -m aquacrop.solution.growing_degree_day
python -m aquacrop.solution.drainage
python -m aquacrop.solution.rainfall_partition
python -m aquacrop.solution.check_groundwater_table
python -m aquacrop.solution.soil_evaporation
python -m aquacrop.solution.root_development
python -m aquacrop.solution.infiltration
python -m aquacrop.solution.HIref_current_day
python -m aquacrop.solution.biomass_accumulation